# Custom elements

## Introduction

The Custom Elements module provides the framework for rendering Drupal data
(entities, fields, ...) into custom elements markup. Custom elements can be
easily rendered by frontend components, e.g. via web components or various
Javascript frontend frameworks. This enables Drupal to render into high-level
theme components, while the actually rendering of the components is handled by
a frontend application (possibly in the browser).

The Custom Elements module provides
 * the API to build a (nested tree) of custom element objects, with associated
cache metadata
 * the API to serialize a tree of custom objects into markup or into
a JSON representation
 * the API for other modules to customize how data is rendered into custom
elements via Custom element processors

### Frontend rendering

Today's browsers provide [an API](https://html.spec.whatwg.org/multipage/custom-elements.html#custom-elements-autonomous-example)
for developers to define their own HTML elements, like
`<flag-icon country="nl"></flag-icon>`. Besides that, many frontend frameworks
render their components using the same, or similar custom elements syntax.
That way, we can render a custom element with [Web components](https://developer.mozilla.org/de/docs/Web/Web_Components) or suiting frontend
frameworks, like [Vue.js](https://vuejs.org/).

## Usage

### Progressive decoupling

Custom elements output may be used as part of a regular Drupal-rendered
response, where some JavaScript or Web-Components take over the rendering in
the browser. For that use case, any Javascript libraries needed for rendering
the markup may be added to the `custom_elements/main` library, e.g. in a custom
theme. This library is attached to every custom element markup rendered via
Drupal.

To use custom elements for outputting specific entity bundles / view modes,
enable the corresponding checkbox on their "Manage Display" tab.

### Full decoupling

Alternatively, the module can be used in conjunction via the Lupus Custom Elements Renderer
modules or Lupus Decoupled Drupal. The [Lupus Custom Elements Renderer](https://www.drupal.org/project/lupus_ce_renderer)
modules switches Drupal's main content renderer to provide API responses using
custom elements markup or a custom elements JSON serialization for complete
pages.

When such a module calls the Custom Elements API directly, the explicit
configuration on the "Manage display" tab is unnecessary.

Lupus Decoupled Drupal builds upon Lupus Custom Elements Renderer and packages
it up into an easily usable decoupled setup, see https://lupus-decoupled.org

## Methods of rendering custom elements

By default, the module does nothing unless its API is used by other modules
(e.g. the Lupus CE renderer) or it's configured to take over entity rendering
for some entity view modes (per "Progressive decoupling" above).

An entity-view can be rendered into custom elements by one of three methods:

1. Using layouts

This method renders a layout, as configured with Drupal's layout builder, into
`<drupal-layout>` elements and allows the contained blocks to render into
custom elements.

This method is used when two conditions are both met:
* "Use layout" is enabled for the requested view mode's Custom Elements Display
  (or for the default display if none exists for the requested view mode);
* "Use Layout Builder" is enabled for the requested view mode's display (or for
  the default display if none exists for the requested view mode). If "Use
  Layout Builder" is not enabled here, the setting on the Custom Elements
  Display is ignored, and one of the other two methods is used.

2. Using Custom Elements Displays

A Custom Elements Display contains information to render individual fields
using 'custom element field formatter plugins'; values are formatted and output
as configured by each plugin. This is equivalent to field formatters used by
Drupal's regular displays.

The Custom Elements UI submodule may be enabled to configure Custom Elements
Displays via the UI.

If no Custom Elements Displays are configured yet for a certain entity/bundle,
a default display for the "Default" view mode is created internally and used
for rendering. This is equivalent to how Drupal's regular displays work. The
same fields are enabled as in Drupal's regular 'internal default' display, and
all fields are rendered in "Automatic processing" mode.

3. Using automatic processing (PHP code)

This method uses processor services (PHP classes) to render entities. This was
the default in version 2.x and is still used when "Use automatic processing"
is enabled in the Custom Elements display.

Processor services can also be used to render individual fields; this is still
the default (i.e. a newly created Custom Elements Display has all fields set
to the "Automatic processing" formatter.)

See "Automatic rendering with processors" for more information below.

## Custom element render formats

The module supports two main serialization modes:

* Markup - Render custom elements into HTML-like markup.
* JSON - Render a JSON tree representation of custom elements.

The markup variant supports multiple styles to cater for the different needs
of different frameworks.

### Custom Element markup styles

Custom elements use "slots" for handling content distribution, i.e. for passing
nested content to an element. However, the concrete syntax used for handling
slots may differ by various frameworks. Thus, the module supports rendering to
different markup styles while it defaults to the Web component style syntax,
which is supported by Vue 2 as well (via its legacy slot syntax). In addition,
the module supports the [Vue2 and Vue 3](https://vuejs.org/v2/guide/components-slots.html#Named-Slots-Shorthand)
syntax which can be enabled via config:

    drush config:set custom_elements.settings  markup_style vue-3

## Further details

### Automatic rendering with processors

The module comes with Custom element processors that try to implement
a reasonable default for content entities and fields. These defaults can be
further customized with custom modules as suiting.

The default entity processor renders the data of all visible fields either as
attribute to the custom element tag, or as nested markup via a slot. The module
maps simple fields and their properties to attributes and falls back to
rendering more complex fields to regular markup, which gets added as slot to
the parent custom element.

## Upgrade from 2.x

**Note** that update hook was added to Custom elements v3 that automatically
enables Custom elements UI module. Read the next section for implications of
this upgrade.

### Preparation

Make sure some submodules are not enabled, or deal with uninstalling the
nonexistent modules later: (They have been removed as of the first 3.x release
because of their very small install base.)

* custom_elements_everywhere:
  There is no 100% replacement; all entities/bundles can have Custom Elements
  enabled in the corresponding "Manage display" tab in the UI. (In terms of
  configuration: in each applicable
  core.entity_view_display.TYPE.BUNDLE.VIEWMODE configuration entity, set
  third_party_settings.custom_elements.enabled = true.)

* custom_elements_thunder:
  Version 3.x ships without any of the processors that were present in 8.x-2.x.
  All these processors applied to a paragraph type; it should be possible to
  replace these with Custom Elements Displays (configuration) that are now
  shipped with the submodule. See
  [the module's README.md](modules/custom_elements_thunder/README.md).

### Do a test upgrade first, then plan your own update path.

The update introduces breaking changes in the way entities are rendered, and
there is no automatic update path. The reason for this is:

* There is no 'registry' of bundles and view modes which your site uses for
  rendering custom entities. You could even be rendering entities in view
  modes with names that are unknown to Drupal.
* There's a way to provide almost-perfect backward compatibility, but it
  creates unnecessary configuration that is likely detrimental in the longer
  term.

To have entities be rendered by the exact same processors as in version 2,
create a Custom Elements Display for the default view mode of all bundles of
all entity types in your site, that you use to render custom elements. If you
don't know which bundles these are, then you could create Custom Elements
Displays or all bundles. See below code that could be used in an update hook:

**TO DO** - this example should be updated for core entity view displays that
have layout builder enabled.
```php
// Note v3 must be installed already.
$storage = \Drupal::entityTypeManager()->getStorage('entity_ce_display');
foreach ($ALL_APPROPRIATE_ENTITY_TYPES as $entity_type_id) {
    /** @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface $bundle_info */
    $bundle_info = \Drupal::service('entity_type.bundle.info');
    $bundles = $bundle_info->getBundleInfo($entity_type_id);
    foreach (array_keys($bundles) as $bundle_key) {
      // Skip already existing CE displays.
      $ce_display_id = "$entity_type_id.$bundle_key.default";
      if ($storage->load($ce_display_id) === NULL) {
        // Create and save CE display with default values (it gets roughly the
        // same enabled fields as a default Core entity view display), but
        // turn on automatic processing. (This means those fields aren't used
        // at all and do not show in the UI. Already enabling these fields can
        // provide a marginal benefit for later editing, though.)
        EntityCeDisplay::create([
          'targetEntityType' => $entity_type_id,
          'bundle' => $bundle_key,
          'mode' => 'default',
        ])->setForceAutoProcessing(TRUE)
          ->save();
      }
    }
  }
}
```

This means all view modes (including future new ones) that will be used for
rendering custom elements, will fall back to "Automatic processing". Many site
owners will instead prefer having explicitly configured Custom Elements
Displays for all the view modes they use, to provide more flexibility
(configurability through the UI) in the future. Therefore, the above code is
not included in a standard update hook.

Also, bundles which are rendered using the standard
DefaultContententityProcessor in version 2, very likely do not need
"Automatic processing" enabled, to be backward compatible: the default settings
of Custom Elements Displays produce exactly the same output.

Please weigh the options of
* enabling "Automatic processing" everywhere, during the upgrade, per above code
* investigating which exact bundles + view modes you use for rendering custom
  elements, and only creating Custom Elements Displays with "Automatic
  processing" for those
* investigating the same, and testing if your output is the same in the default
  situation and/or if you can tweak Custom Elements Displays slightly, export
  the configuration, and re-import them during/after the production update.

## Credits

  - [drunomics GmbH](https://www.drupal.org/drunomics): Concept, Development, Maintenance
  - [Österreichischer Wirtschaftsverlag GmbH](https://www.drupal.org/%C3%B6sterreichischer-wirtschaftsverlag-gmbh): Initial sponsor of v1
