<?php

namespace Drupal\custom_elements;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\Entity\EntityViewDisplay;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Render\Element;
use Drupal\Core\Session\AccountInterface;
use Drupal\custom_elements\Entity\EntityCeDisplayInterface;
use Drupal\custom_elements\Processor\CustomElementProcessorInterface;

/**
 * Service to preprocess template variables for custom elements.
 */
class CustomElementGenerator {

  use CustomElementsBlockRenderHelperTrait;
  use CustomElementsProcessorFieldUtilsTrait;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Entity repository.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Array of all processors and their priority.
   *
   * @var array
   */
  protected $processorByPriority = [];

  /**
   * Sorted list of registered processors.
   *
   * @var \Drupal\custom_elements\Processor\CustomElementProcessorInterface[]
   */
  protected $sortedProcessors;

  /**
   * CustomElementGenerator constructor.
   *
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler.
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(ModuleHandlerInterface $moduleHandler, EntityRepositoryInterface $entity_repository, EntityTypeManagerInterface $entity_type_manager) {
    $this->moduleHandler = $moduleHandler;
    $this->entityRepository = $entity_repository;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Adds a processor.
   *
   * @param \Drupal\custom_elements\Processor\CustomElementProcessorInterface $processor
   *   The processor to add.
   * @param int $priority
   *   The priority for the processor.
   */
  public function addProcessor(CustomElementProcessorInterface $processor, $priority = 0) {
    $this->processorByPriority[$priority][] = $processor;
    // Force the processors to be re-sorted.
    $this->sortedProcessors = NULL;
  }

  /**
   * Gets an array of processors, sorted by their priority.
   *
   * @return \Drupal\custom_elements\Processor\CustomElementProcessorInterface[]
   *   Returns sorted processors.
   */
  public function getSortedProcessors() {
    if (!isset($this->sortedProcessors)) {
      // Sort the processors according to priority.
      krsort($this->processorByPriority);

      // Merge nested processors from $this->processors
      // into $this->sortedProviders.
      $this->sortedProcessors = [];
      foreach ($this->processorByPriority as $processors) {
        $this->sortedProcessors = array_merge($this->sortedProcessors, $processors);
      }
    }
    return $this->sortedProcessors;
  }

  /**
   * Gets the defaults to apply for the given entity type, bundle and view mode.
   *
   * @return \Drupal\custom_elements\CustomElement
   *   A new custom element with the defaults set.
   */
  public function getViewModeDefaults(string $entityType, string $bundle, string $viewMode) {
    $bundle_key = $this->entityTypeManager->getDefinition($entityType)
      ->getKey('bundle');
    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $entity = $this->entityTypeManager->getStorage($entityType)
      ->create($bundle_key && $bundle ? [$bundle_key => $bundle] : []);

    return $this->getEntityDefaults($entity, $viewMode);
  }

  /**
   * Gets the defaults to apply for the given entity and view mode.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity.
   * @param string $viewMode
   *   The view mode.
   *
   * @return \Drupal\custom_elements\CustomElement
   *   A new custom element with the defaults set.
   */
  public function getEntityDefaults(ContentEntityInterface $entity, $viewMode) {
    $custom_element = new CustomElement();
    $custom_element->addCacheableDependency($entity);

    // By default output tags like {entity}-{bundle}-{view_mode} or
    // {entity}-{view_mode} if the entity type does not make use of
    // bundles.
    $tag = $entity->getEntityTypeId();
    if (!empty($entity->getEntityType()->getKeys()['bundle'])) {
      $tag .= '-' . $entity->bundle();
    }
    if ($viewMode != 'default') {
      $tag .= '-' . $viewMode;
    }
    $custom_element->setTag($tag);
    $custom_element->setTagPrefix('');

    // Allow altering the element for the given before and after.
    $this->moduleHandler->alter('custom_element_entity_defaults', $custom_element, $entity, $viewMode);
    return $custom_element;
  }

  /**
   * Generates a custom element tag for the given entity and view mode.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Entity to process.
   * @param string $viewMode
   *   View mode used for rendering field values into slots.
   * @param string|null $langcode
   *   (optional) For which language the entity should be rendered, defaults to
   *   the current content language.
   *
   * @return \Drupal\custom_elements\CustomElement
   *   Extracted custom elements containing data attributes and slots.
   */
  public function generate(ContentEntityInterface $entity, string $viewMode, string $langcode = NULL) {
    // Get desired entity translation.
    $entity = $this->entityRepository->getTranslationFromContext($entity, $langcode);
    $custom_element = $this->getEntityDefaults($entity, $viewMode);
    $this->buildEntityContent($entity, $custom_element, $viewMode);
    $this->moduleHandler->alter('custom_element_entity', $custom_element, $entity, $viewMode);
    return $custom_element;
  }

  /**
   * Processes the given data and adds it to the custom element.
   *
   * @param mixed $data
   *   The data.
   * @param \Drupal\custom_elements\CustomElement $custom_element
   *   The custom element to which to add it.
   * @param string $viewMode
   *   The current view mode.
   * @param string $key
   *   (Optional) Name to use for adding data into (a slot / attribute in) the
   *   custom element. Processors can use this how they see fit. Most use it
   *   as a literal key value; some may use it as e.g. a key prefix to add
   *   several values.
   */
  public function process($data, CustomElement $custom_element, $viewMode, $key = '') {
    foreach ($this->getSortedProcessors() as $processor) {
      if ($processor->supports($data, $viewMode)) {
        $processor->addtoElement($data, $custom_element, $viewMode, $key);
        break;
      }
    }
  }

  /**
   * Builds an entity's content as custom element.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity.
   * @param \Drupal\custom_elements\CustomElement $element
   *   The custom element to add to.
   * @param string $viewMode
   *   The current view mode.
   */
  public function buildEntityContent(ContentEntityInterface $entity, CustomElement $element, $viewMode) {
    $entity_ce_display = $this->getEntityCeDisplay($entity->getEntityTypeId(), $entity->bundle(), $viewMode);
    $this->buildCeDisplay($entity, $element, $entity_ce_display, $viewMode);
  }

  /**
   * Builds an entity's content a custom element using its layout's regions.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param \Drupal\custom_elements\CustomElement $custom_element
   *   The custom element.
   * @param \Drupal\custom_elements\CustomElementsLayoutBuilderEntityViewDisplay $display
   *   The view display of the current view mode.
   */
  protected function buildLayoutBuilderContent(EntityInterface $entity, CustomElement $custom_element, CustomElementsLayoutBuilderEntityViewDisplay $display) {
    $section_elements = [];
    $build = $display->buildLayoutSections($entity);

    // Loop over sections and convert render array back to custom elements
    // if blocks render into custom elements.
    foreach (Element::children($build, TRUE) as $key) {
      $section_element = CustomElement::create('drupal-layout');
      $section_build = $build[$key];
      /** @var \Drupal\Core\Layout\LayoutDefinition $layout */
      $layout = $section_build['#layout'];
      $section_element->setAttribute('layout', $layout->id());
      foreach ($layout->getRegions() as $region_name => $region) {
        if (!empty($section_build[$region_name])) {
          $elements = $this->getElementsFromBlockContentRenderArray($section_build[$region_name], $section_element);
          $section_element->addSlotFromNestedElements($region_name, $elements);
        }
      }
      $section_element->setAttribute('settings', $section_build['#settings']);
      $section_elements[] = $section_element;
    }

    $custom_element->setSlotFromNestedElements('sections', $section_elements);
    $custom_element->addCacheableDependency(BubbleableMetadata::createFromRenderArray($build));
    $custom_element->addCacheableDependency($display);
  }

  /**
   * Builds an entity's content as custom element using a CE display.
   *
   * Not accessible or empty fields are not built.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity.
   * @param \Drupal\custom_elements\CustomElement $custom_element
   *   The custom element.
   * @param \Drupal\custom_elements\Entity\EntityCeDisplayInterface $display
   *   Entity custom element display entity.
   * @param string $viewMode
   *   The view mode being built.
   * @param \Drupal\Core\Session\AccountInterface|null $account
   *   (optional) The user for which to check access, or NULL to check access
   *   for the current user. Defaults to NULL.
   */
  public function buildCeDisplay(ContentEntityInterface $entity, CustomElement $custom_element, EntityCeDisplayInterface $display, string $viewMode, AccountInterface $account = NULL) {
    $custom_element->addCacheableDependency($display);

    // If 'use layout builder' is enabled, and if it's enabled on the
    // corresponding core display (for view mode or default), skip processing
    // of the fields in the CE display, and let layout builder render it all.
    if ($display->getUseLayoutBuilder()) {
      $entity_view_displays = EntityViewDisplay::collectRenderDisplays([$entity], $viewMode);
      $entity_view_display = reset($entity_view_displays);
      if ($entity_view_display->getThirdPartySetting('layout_builder', 'enabled')) {
        assert($entity_view_display instanceof CustomElementsLayoutBuilderEntityViewDisplay);
        $this->buildLayoutBuilderContent($entity, $custom_element, $entity_view_display);
        return;
      }
    }

    $display->setOriginalMode($viewMode);

    $custom_element->setTag($display->getCustomElementName());

    // When configured, only do auto-processing on entity-level and skip the
    // config of individual components. This is the same behaviour as it was
    // default in 2.x.
    if ($display->getForceAutoProcessing()) {
      $this->process($entity, $custom_element, $viewMode);
      return;
    }
    // Else render using the individual display components.
    foreach ($display->getComponents() as $component_name => $display_component) {
      if ($formatter = $display->getRenderer($component_name)) {
        $field_name = $display_component['field_name'];
        if ($this->fieldIsAccessible($entity, $field_name, $custom_element, $account)) {
          // @todo Move prepareBuild to prepareView phase.
          $formatter->prepareBuild([$entity->get($field_name)]);
          $formatter->build($entity->get($field_name), $custom_element);
        }
      }
    }
  }

  /**
   * Gets a Custom Elements display for the entity type, bundle and view mode.
   *
   * @param string $entityTypeId
   *   The entity type.
   * @param string $bundle
   *   The bundle.
   * @param string $viewMode
   *   The view mode.
   *
   * @return \Drupal\custom_elements\Entity\EntityCeDisplayInterface
   *   The applicable CE display. If none exists / is enabled, a suiting
   *   display is auto-generated.
   */
  public function getEntityCeDisplay(string $entityTypeId, string $bundle, string $viewMode) {
    /** @var \Drupal\custom_elements\Entity\EntityCeDisplayInterface $entity_ce_display */
    $entity_ce_display = $this->entityTypeManager
      ->getStorage('entity_ce_display')
      ->load("$entityTypeId.$bundle.$viewMode");

    if (!$entity_ce_display || !$entity_ce_display->status()) {
      // Fall back to default view mode. If it's not in use, return NULL.
      $entity_ce_display = $this->entityTypeManager
        ->getStorage('entity_ce_display')
        ->load("$entityTypeId.$bundle.default");
      if (!$entity_ce_display || !$entity_ce_display->status()) {
        // No usable default, so auto-create a suiting entity.
        $entity_ce_display = $this->entityTypeManager->getStorage('entity_ce_display')->create([
          'id' => "$entityTypeId.$bundle.$viewMode",
          'targetEntityType' => $entityTypeId,
          'bundle' => $bundle,
          'mode' => $viewMode,
          'status' => TRUE,
        ]);
      }
    }
    return $entity_ce_display;
  }

}
