<?php

namespace Drupal\custom_elements\Plugin\CustomElementsFieldFormatter;

use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\custom_elements\CustomElement;
use Drupal\custom_elements\CustomElementsFieldFormatterBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Implementation of the 'video_embed' custom element formatter plugin.
 *
 * Sets embed URL and thumbnail from the applicable video provider.
 *
 * @CustomElementsFieldFormatter(
 *   id = "video_embed",
 *   label = @Translation("Video embed URL + thumbnail"),
 *   field_types = {
 *     "video_embed_field"
 *   },
 *   weight = -10
 * )
 */
class VideoEmbedCeFieldFormatter extends CustomElementsFieldFormatterBase {

  /**
   * Video embed provider manager.
   *
   * @var \Drupal\video_embed_field\ProviderManager
   */
  protected $providerManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->providerManager = $container->get('video_embed_field.provider_manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function build(FieldItemListInterface $items, CustomElement $custom_element, $langcode = NULL) {
    // Process single field value only, because setting multiple properties
    // (of which one always in an attribute) does not make sense with multiple
    // values.
    $field_item = $items->first();
    $value = $field_item->value;
    $provider = $this->providerManager->loadProviderFromInput($value);
    if (!$provider) {
      return;
    }
    $element_name = $this->getName() ?: 'src';

    $embed_code = $provider->renderEmbedCode('0', '0', FALSE);
    if ($this->isSlot()) {
      $custom_element->addSlot($element_name, $embed_code['#url']);
    }
    else {
      $custom_element->setAttribute($element_name, $embed_code['#url']);
    }

    $preview_image_src = $provider->getRemoteThumbnailUrl();
    if (!empty($this->configuration['thumbnail_force_https'])) {
      $preview_image_src = str_replace('http:', 'https:', $preview_image_src);
    }
    $custom_element->setAttribute($element_name . '-thumbnail', $preview_image_src);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return parent::defaultConfiguration() + ['thumbnail_force_https' => FALSE];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['thumbnail_force_https'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Force HTTPS for thumbnail URL'),
      '#default_value' => $this->configuration['thumbnail_force_https'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['thumbnail_force_https'] = $form_state->getValue('thumbnail_force_https');
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];
    if ($this->getFieldDefinition()->getFieldStorageDefinition()->getCardinality() != 1) {
      $summary[] = $this->t('Only the first field value is output.');
    }
    if (!empty($this->configuration['thumbnail_force_https'])) {
      $summary[] = $this->t('Force HTTPS for thumbnail URL.');
    }
    return $summary;
  }

}
