<?php

namespace Drupal\Tests\custom_elements\Kernel;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\Entity\EntityViewMode;
use Drupal\custom_elements\CustomElementGeneratorTrait;
use Drupal\custom_elements\Entity\EntityCeDisplay;
use Drupal\custom_elements\Entity\EntityCeDisplayInterface;
use Drupal\entity_test\Entity\EntityTest;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\node\Traits\NodeCreationTrait;

/**
 * Tests rendering of entityreference fields.
 *
 * @group custom_elements
 */
class CustomElementGeneratorTest extends KernelTestBase {

  use CustomElementGeneratorTrait;
  use NodeCreationTrait;

  /**
   * {@inheritdoc}
   *
   * @var bool
   * @todo Fix config schema for CE-displays and re-enable.
   */
  protected $strictConfigSchema = FALSE;

  /**
   * Default CE display, initialized/used by individual tests.
   */
  protected EntityCeDisplayInterface $defaultCeDisplay;

  /**
   * Full CE display, initialized/used by individual tests.
   */
  protected EntityCeDisplayInterface $fullCeDisplay;

  /**
   * Entity, initialized/used by individual tests.
   */
  protected ContentEntityInterface $testEntity;

  /**
   * View modes, initialized/used by individual tests.
   */
  protected array $coreViewModes = [];

  /**
   * Modules to install.
   *
   * @var string[]
   */
  protected static $modules = [
    'custom_elements',
    'entity_test',
    'field',
    'text',
    'block',
    'user',
    'layout_discovery',
    'layout_builder',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    $this->installEntitySchema('user');
  }

  /**
   * Tests that getEntityCeDisplay() takes the default CE display as fallback.
   */
  public function testViewModeFallback() {
    EntityViewMode::create([
      'id' => 'entity_test.full',
      'targetEntityType' => 'entity_test',
    ])->save();
    $ceDisplayDefault = EntityCeDisplay::create([
      'targetEntityType' => 'entity_test',
      'customElementName' => 'ce-test',
      'bundle' => 'entity_test',
      'mode' => 'default',
    ]);
    $ceDisplayDefault->save();
    $loadedDisplay = $this->getCustomElementGenerator()->getEntityCeDisplay('entity_test', 'entity_test', 'full');
    $this->assertSame($ceDisplayDefault->getCustomElementName(), $loadedDisplay->getCustomElementName());

    $ceDisplayFull = EntityCeDisplay::create([
      'targetEntityType' => 'entity_test',
      'customElementName' => 'ce-test-full',
      'bundle' => 'entity_test',
      'mode' => 'full',
    ]);
    $ceDisplayFull->save();
    $loadedDisplay = $this->getCustomElementGenerator()->getEntityCeDisplay('entity_test', 'entity_test', 'full');
    $this->assertSame($ceDisplayFull->getCustomElementName(), $loadedDisplay->getCustomElementName());

    $ceDisplayFull->setStatus(FALSE)->save();
    $loadedDisplay = $this->getCustomElementGenerator()->getEntityCeDisplay('entity_test', 'entity_test', 'full');
    $this->assertSame($ceDisplayDefault->getCustomElementName(), $loadedDisplay->getCustomElementName());

    // When no CE displays are active, one is created on the fly; the default
    // name is derived from the entity type/bundle.
    $ceDisplayDefault->setStatus(FALSE)->save();
    $loadedDisplay = $this->getCustomElementGenerator()->getEntityCeDisplay('entity_test', 'entity_test', 'full');
    $this->assertSame('entity-test-entity-test-full', $loadedDisplay->getCustomElementName());
  }

  /**
   * Tests building custom elements using generate().
   */
  public function testGenerate() {
    // With [array-key-part 1] CE displays enabled, build custom element in
    // [array-key-part 2] view mode. Expect [value] to be used for generating.
    $this->dotestGenerate([], [
      'default:default' => 'ce-default',
      'default:full'    => 'ce-default',
      'both:default'    => 'ce-default',
      'both:full'       => 'ce-full',

      // neither: and full: are not very realistic; they imply that the default
      // CE display exists and is explicitly disabled (which cannot be done
      // through the UI). Test only once, skip in further tests: we can assume
      // that full:full == both:full always, and the 3 others never change.
      'neither:default' => 'autocreated-ce-default',
      'neither:full'    => 'autocreated-ce-full',
      'full:default'    => 'autocreated-ce-default',
      'full:full'       => 'ce-full',
    ]);
  }

  /**
   * Tests generate() with CE displays set to 'auto processing'.
   */
  public function testGenerateAuto() {
    // With [array-key-part 1] CE displays enabled, build CE in
    // [array-key-part 2] view mode. Expect [value] to be used.
    $this->dotestGenerate(['forceAutoProcessing' => TRUE], [
      'default:default' => 'autoprocessed-ce-default',
      'default:full'    => 'autoprocessed-ce-default',
      'both:default'    => 'autoprocessed-ce-default',
      'both:full'       => 'autoprocessed-ce-full',
    ]);
  }

  /**
   * Tests generate() with CE displays set to LB, but no layouts present.
   */
  public function testGenerateLayoutWithoutCoreDisplays() {
    // No core display present (i.e. default view display autogenerted by core):
    // same results as testGenerate() / auto().
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'ce-default',
      'default:full'    => 'ce-default',
      'both:default'    => 'ce-default',
      'both:full'       => 'ce-full',
    ]);

    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'autoprocessed-ce-default',
      'default:full'    => 'autoprocessed-ce-default',
      'both:default'    => 'autoprocessed-ce-default',
      'both:full'       => 'autoprocessed-ce-full',
    ]);
  }

  /**
   * Tests generate() with CE displays set to LB, and core displays without LB.
   *
   * Difference with previous test: a 'full' display is present, but it has
   * no LB setting. This should have no effect -> give the same result.
   */
  public function testGenerateLayoutWithCoreDisplaysWithoutLayout() {
    $this->prepareTestGenerateCoreDisplays(['default*', 'full*']);
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'ce-default',
      'default:full'    => 'ce-default',
      'both:default'    => 'ce-default',
      'both:full'       => 'ce-full',
    ]);

    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'autoprocessed-ce-default',
      'default:full'    => 'autoprocessed-ce-default',
      'both:default'    => 'autoprocessed-ce-default',
      'both:full'       => 'autoprocessed-ce-full',
    ]);
  }

  /**
   * Tests generate() with CE set to LB, and both core displays having LB.
   */
  public function testGenerateLayoutWithBothCoreDisplays() {
    $this->prepareTestGenerateCoreDisplays(['default', 'full']);
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'layout',
      // Fall back to default CE display, see LB enabled, use layout from full
      // core view display.
      'default:full'    => 'layout',
      'both:default'    => 'layout',
      'both:full'       => 'layout',
    ]);

    // forceAutoProcessing makes no difference: useLayoutBuilder has
    // precedence.
    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'layout',
      'default:full'    => 'layout',
      'both:default'    => 'layout',
      'both:full'       => 'layout',
    ]);
  }

  /**
   * Tests generate() with CE set to LB, and default core display having LB.
   */
  public function testGenerateLayoutWithDefaultCoreDisplay() {
    $this->prepareTestGenerateCoreDisplays(['default']);
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'layout',
      // Fall back to default CE display, see LB enabled, use layout from
      // default core view display.
      'default:full'    => 'layout',
      'both:default'    => 'layout',
      'both:full'       => 'layout',
    ]);

    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'layout',
      'default:full'    => 'layout',
      'both:default'    => 'layout',
      'both:full'       => 'layout',
    ]);
  }

  /**
   * Tests generate() with CE set to LB, and full core display having LB.
   */
  public function testGenerateLayoutWithFullCoreDisplay() {
    $this->prepareTestGenerateCoreDisplays(['full']);
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'ce-default',
      // Fall back to default CE display, see LB enabled, use layout from full
      // core view display.
      'default:full'    => 'layout',
      'both:default'    => 'ce-default',
      'both:full'       => 'layout',
    ]);

    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'autoprocessed-ce-default',
      'default:full'    => 'layout',
      'both:default'    => 'autoprocessed-ce-default',
      'both:full'       => 'layout',
    ]);
  }

  /**
   * Tests generate() with CE set to LB, default display with LB, full without.
   */
  public function testGenerateLayoutWithDefaultCoreDisplayAndFullWithout() {
    $this->prepareTestGenerateCoreDisplays(['default', 'full*']);
    $this->dotestGenerate(['useLayoutBuilder' => TRUE], [
      'default:default' => 'layout',
      // Fall back to default CE display, see LB enabled, check full core
      // display, do not see LB enabled, fall back to building using CE display.
      'default:full'    => 'ce-default',
      'both:default'    => 'layout',
      'both:full'       => 'ce-full',
    ]);

    $this->dotestGenerate(['useLayoutBuilder' => TRUE, 'forceAutoProcessing' => TRUE], [
      'default:default' => 'layout',
      'default:full'    => 'autoprocessed-ce-default',
      'both:default'    => 'layout',
      'both:full'       => 'autoprocessed-ce-full',
    ]);
  }

  /**
   * Prepares doTestGenerate(), can be called only once per test.
   *
   * @param string[] $viewModes
   *   View modes to create core displays for; ending in '*' if the display
   *   should not have Layout Builder enabled.
   */
  protected function prepareTestGenerateCoreDisplays(array $viewModes) {
    // View modes can be created only after the view mode is created, so
    // store in class variable. This is a separate method only as a reminder
    // that these core displays can't be created for different view modes, when
    // calling doTestGenerate() repeatedly.
    $this->coreViewModes = $viewModes;
  }

  /**
   * Perform tests on generate() with set conditions.
   *
   * If core display view modes are needed, prepareTestGenerate() must be
   * called beforehand.
   *
   * @param array $ceDisplayAddConfig
   *   Configuration values to set for CE display, in addition to standard
   *   values as set in this method.
   * @param array $conditions
   *   An array with expected results that signify which display was used for
   *   generating a custom element, having a fixed set of required keys that
   *   describe conditions. Examples:
   *   'both:full':    with both default & full displays enabled, build full.
   *   'default:full': with only default display enabled, build full display.
   */
  protected function doTestGenerate(array $ceDisplayAddConfig, array $conditions): void {
    if (isset($this->fullCeDisplay)) {
      // Do not reinitialize fields / core displays; delete CE displays so they
      // are reinitialized.
      $this->fullCeDisplay->delete();
      $this->defaultCeDisplay->delete();
    }
    else {
      FieldStorageConfig::create([
        'entity_type' => 'entity_test',
        'field_name' => 'field_string',
        'type' => 'string',
        'cardinality' => 1,
      ])->save();
      FieldConfig::create([
        'entity_type' => 'entity_test',
        'field_name' => 'field_string',
        'bundle' => 'entity_test',
        'label' => 'Test string-field',
        'translatable' => FALSE,
      ])->save();
      $this->testEntity = EntityTest::create([
        'type' => 'entity_test',
        'name' => 'Test Entity',
        'field_string' => 'This is a test',
      ]);

      EntityViewMode::create([
        'id' => 'entity_test.full',
        'targetEntityType' => 'entity_test',
      ])->save();

      foreach ($this->coreViewModes as $viewMode) {
        $layoutBuilderDisabled = str_ends_with($viewMode, '*');
        if ($layoutBuilderDisabled) {
          $viewMode = substr($viewMode, 0, strlen($viewMode) - 1);
        }
        /** @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface $coreDisplayRepository */
        $coreDisplayRepository = $this->container->get('entity_display.repository');
        $display = $coreDisplayRepository->getViewDisplay('entity_test', 'entity_test', $viewMode);
        if (!$layoutBuilderDisabled) {
          $display->setThirdPartySetting('layout_builder', 'enabled', TRUE);
        }
        $display->save();
      }
    }
    // Create CE display with field_string enabled; autocreated displays do not
    // have that.
    $this->defaultCeDisplay = EntityCeDisplay::create($ceDisplayAddConfig + [
      'targetEntityType' => 'entity_test',
      'customElementName' => 'ce-test',
      'bundle' => 'entity_test',
      'mode' => 'default',
    ])
      ->removeComponent('name')
      ->setComponent('text', ['field_name' => 'field_string', 'formatter' => 'auto']);
    $this->defaultCeDisplay->save();
    $this->doTestGenerateAssertions('default', $conditions['default:default']);
    $this->doTestGenerateAssertions('full', $conditions['default:full']);

    $this->fullCeDisplay = EntityCeDisplay::create($ceDisplayAddConfig + [
      'targetEntityType' => 'entity_test',
      'customElementName' => 'ce-test-full',
      'bundle' => 'entity_test',
      'mode' => 'full',
    ])
      ->setComponent('text', ['field_name' => 'field_string', 'formatter' => 'auto']);
    $this->fullCeDisplay->save();
    $this->doTestGenerateAssertions('default', $conditions['both:default']);
    $this->doTestGenerateAssertions('full', $conditions['both:full']);

    if (isset($conditions['full:default']) || isset($conditions['full:full'])) {
      $this->defaultCeDisplay->setStatus(FALSE)->save();
      if (isset($conditions['full:default'])) {
        $this->doTestGenerateAssertions('default', $conditions['full:default']);
      }
      if (isset($conditions['full:full'])) {
        $this->doTestGenerateAssertions('full', $conditions['full:full']);
      }
    }

    if (isset($conditions['neither:default']) || isset($conditions['neither:full'])) {
      $this->fullCeDisplay->setStatus(FALSE)->save();
      if (isset($conditions['neither:default'])) {
        $this->doTestGenerateAssertions('default', $conditions['neither:default']);
      }
      if (isset($conditions['neither:full'])) {
        $this->doTestGenerateAssertions('full', $conditions['neither:full']);
      }
    }
  }

  /**
   * Checks which display was used for building an entity into a custom element.
   */
  protected function doTestGenerateAssertions(string $viewMode, string $expectedDisplayUsage): void {
    $customElement = $this->getCustomElementGenerator()->generate($this->testEntity, $viewMode);

    $expectedTagWithAutoDisplay = '';

    // The actual asserts are abstracted away into here, because they are not
    // very interesting. They work only for the test entity and displays as
    // exactly created/configured by doTestGenerate().
    switch ($expectedDisplayUsage) {
      case 'ce-default':
        $this->assertSame($this->defaultCeDisplay->getCustomElementName(), $customElement->getTag());
        $this->assertNull($customElement->getAttribute('name'));
        $this->assertSame('This is a test', $customElement->getAttribute('text'));
        $this->assertEmpty($customElement->getSlots());
        break;

      case 'ce-full':
        $this->assertSame($this->fullCeDisplay->getCustomElementName(), $customElement->getTag());
        $this->assertSame('Test Entity', $customElement->getAttribute('name'));
        $this->assertSame('This is a test', $customElement->getAttribute('text'));
        $this->assertEmpty($customElement->getSlots());
        break;

      case 'autoprocessed-ce-default':
        // Enabled fields are determined from Core display; tag is still set
        // from CE display.
        $expectedTagWithAutoDisplay = $this->defaultCeDisplay->getCustomElementName();
        break;

      case 'autoprocessed-ce-full':
        $expectedTagWithAutoDisplay = $this->fullCeDisplay->getCustomElementName();
        break;

      case 'autocreated-ce-default':
        // CE display is auto-created with fields exactly like Core display;
        // tag is derived from entity ID.
        $expectedTagWithAutoDisplay = 'entity-test-entity-test';
        break;

      case 'autocreated-ce-full':
        $expectedTagWithAutoDisplay = 'entity-test-entity-test-full';
        break;

      case 'layout':
        // Tests do not distinguish yet between the default vs full layout.
        $this->assertNull($customElement->getAttribute('name'));
        $this->assertNull($customElement->getAttribute('text'));
        $this->assertArrayHasKey('sections', $customElement->getSlots());
        break;

      default:
        throw new \LogicException("doTestGenerateAssertions() must define $expectedDisplayUsage.");
    }

    if ($expectedTagWithAutoDisplay) {
      // Check for fields being equal to what the standard Core display has.
      $this->assertSame($expectedTagWithAutoDisplay, $customElement->getTag());
      $this->assertSame('Test Entity', $customElement->getAttribute('name'));
      $this->assertEmpty($customElement->getSlots());
    }
  }

}
